@@ -168,7 +168,8 @@ remote_desktop_dialog_done (GtkWidget                 *widget,
     {
     default:
       g_warning ("Unexpected response: %d", dialog_response);
-      /* Fall through */
+      G_GNUC_FALLTHROUGH;
+
     case GTK_RESPONSE_DELETE_EVENT:
       response = 2;
       break;
