@@ -33,8 +33,15 @@
 #include "externalwindow.h"
 #include "request.h"
 #include "session.h"
+#include "shellintrospect.h"
 #include "utils.h"
 
+#define RESTORE_FORMAT_VERSION 1
+#define RESTORE_VARIANT_TYPE "(xxa(uuv))"
+#define MONITOR_TYPE "s"
+#define WINDOW_TYPE "(ss)"
+#define VIRTUAL_TYPE "b"
+
 typedef struct _ScreenCastDialogHandle ScreenCastDialogHandle;
 
 typedef struct _ScreenCastSession
@@ -49,6 +56,13 @@ typedef struct _ScreenCastSession
 
   ScreenCastSelection select;
 
+  ScreenCastPersistMode persist_mode;
+  GPtrArray *streams_to_restore;
+  struct {
+    GVariant *data;
+    int64_t creation_time;
+  } restored;
+
   GDBusMethodInvocation *start_invocation;
   ScreenCastDialogHandle *dialog_handle;
 } ScreenCastSession;
@@ -77,18 +91,6 @@ static GnomeScreenCast *gnome_screen_cast;
 GType screen_cast_session_get_type (void);
 G_DEFINE_TYPE (ScreenCastSession, screen_cast_session, session_get_type ())
 
-static void
-start_done (ScreenCastSession *session);
-
-static gboolean
-start_session (ScreenCastSession *session,
-               GVariant *selections,
-               GError **error);
-
-static void
-cancel_start_session (ScreenCastSession *session,
-                      int response);
-
 static gboolean
 is_screen_cast_session (Session *session)
 {
@@ -111,10 +113,93 @@ screen_cast_dialog_handle_close (ScreenCastDialogHandle *dialog_handle)
   screen_cast_dialog_handle_free (dialog_handle);
 }
 
+static GVariant *
+serialize_streams_as_restore_data (ScreenCastSession *screen_cast_session,
+                                   GPtrArray         *streams)
+{
+  GVariantBuilder restore_data_builder;
+  GVariantBuilder impl_builder;
+  int64_t creation_time;
+  int64_t last_used_time;
+  guint i;
+
+  if (!streams || streams->len == 0)
+    return NULL;
+
+  last_used_time = g_get_real_time ();
+  if (screen_cast_session->restored.creation_time != -1)
+    creation_time = screen_cast_session->restored.creation_time;
+  else
+    creation_time = g_get_real_time ();
+
+  g_variant_builder_init (&impl_builder, G_VARIANT_TYPE (RESTORE_VARIANT_TYPE));
+  g_variant_builder_add (&impl_builder, "x", creation_time);
+  g_variant_builder_add (&impl_builder, "x", last_used_time);
+
+  g_variant_builder_open (&impl_builder, G_VARIANT_TYPE ("a(uuv)"));
+  for (i = 0; i < streams->len; i++)
+    {
+      ScreenCastStreamInfo *info = g_ptr_array_index (streams, i);
+      GVariant *stream_variant;
+      Monitor *monitor;
+      Window *window;
+
+      switch (info->type)
+        {
+        case SCREEN_CAST_SOURCE_TYPE_MONITOR:
+          monitor = info->data.monitor;
+          stream_variant = g_variant_new (MONITOR_TYPE,
+                                          monitor_get_match_string (monitor));
+          break;
+
+        case SCREEN_CAST_SOURCE_TYPE_WINDOW:
+          window = info->data.window;
+          stream_variant = g_variant_new (WINDOW_TYPE,
+                                          window_get_app_id (window),
+                                          window_get_title (window));
+          break;
+
+        case SCREEN_CAST_SOURCE_TYPE_VIRTUAL:
+          /*
+           * D-Bus doesn't accept maybe types, so just pass bogus boolean. It
+           * doesn't really matter since we'll never actually read this value.
+           */
+          stream_variant = g_variant_new (VIRTUAL_TYPE, TRUE);
+          break;
+        }
+
+      g_variant_builder_add (&impl_builder,
+                             "(uuv)",
+                             info->id,
+                             info->type,
+                             stream_variant);
+    }
+  g_variant_builder_close (&impl_builder);
+
+  g_variant_builder_init (&restore_data_builder, G_VARIANT_TYPE ("(suv)"));
+  g_variant_builder_add (&restore_data_builder, "s", "GNOME");
+  g_variant_builder_add (&restore_data_builder, "u", RESTORE_FORMAT_VERSION);
+  g_variant_builder_add (&restore_data_builder, "v", g_variant_builder_end (&impl_builder));
+  return g_variant_builder_end (&restore_data_builder);
+}
+
+static void
+cancel_start_session (ScreenCastSession *screen_cast_session,
+                      int response)
+{
+  GVariantBuilder results_builder;
+
+  g_variant_builder_init (&results_builder, G_VARIANT_TYPE_VARDICT);
+  xdp_impl_screen_cast_complete_start (XDP_IMPL_SCREEN_CAST (impl),
+                                       screen_cast_session->start_invocation,
+                                       response,
+                                       g_variant_builder_end (&results_builder));
+}
+
 static gboolean
-handle_close (XdpImplRequest *object,
-              GDBusMethodInvocation *invocation,
-              ScreenCastDialogHandle *dialog_handle)
+on_request_handle_close_cb (XdpImplRequest         *object,
+                            GDBusMethodInvocation  *invocation,
+                            ScreenCastDialogHandle *dialog_handle)
 {
   cancel_start_session (dialog_handle->session, 2);
 
@@ -124,10 +209,95 @@ handle_close (XdpImplRequest *object,
 }
 
 static void
-screen_cast_dialog_done (GtkWidget *widget,
-                         int dialog_response,
-                         GVariant *selections,
-                         ScreenCastDialogHandle *dialog_handle)
+on_gnome_screen_cast_session_ready (GnomeScreenCastSession *gnome_screen_cast_session,
+                                    ScreenCastSession      *screen_cast_session)
+{
+  GVariantBuilder streams_builder;
+  GVariantBuilder results_builder;
+
+  g_variant_builder_init (&results_builder, G_VARIANT_TYPE_VARDICT);
+  g_variant_builder_init (&streams_builder, G_VARIANT_TYPE ("a(ua{sv})"));
+
+  gnome_screen_cast_session = screen_cast_session->gnome_screen_cast_session;
+  gnome_screen_cast_session_add_stream_properties (gnome_screen_cast_session,
+                                                   &streams_builder);
+
+  g_variant_builder_add (&results_builder, "{sv}",
+                         "streams",
+                         g_variant_builder_end (&streams_builder));
+
+  if (screen_cast_session->persist_mode != SCREEN_CAST_PERSIST_MODE_NONE)
+    {
+      g_autoptr(GPtrArray) streams = g_steal_pointer (&screen_cast_session->streams_to_restore);
+      GVariant *restore_data;
+
+      restore_data = serialize_streams_as_restore_data (screen_cast_session, streams);
+
+      if (restore_data)
+        {
+          g_variant_builder_add (&results_builder, "{sv}", "persist_mode",
+                                 g_variant_new_uint32 (screen_cast_session->persist_mode));
+          g_variant_builder_add (&results_builder, "{sv}", "restore_data", restore_data);
+        }
+    }
+
+  xdp_impl_screen_cast_complete_start (XDP_IMPL_SCREEN_CAST (impl),
+                                       screen_cast_session->start_invocation, 0,
+                                       g_variant_builder_end (&results_builder));
+  screen_cast_session->start_invocation = NULL;
+}
+
+static void
+on_gnome_screen_cast_session_closed (GnomeScreenCastSession *gnome_screen_cast_session,
+                                     ScreenCastSession      *screen_cast_session)
+{
+  session_close ((Session *)screen_cast_session);
+}
+
+static gboolean
+start_session (ScreenCastSession  *screen_cast_session,
+               GPtrArray          *streams,
+               GError            **error)
+{
+  GnomeScreenCastSession *gnome_screen_cast_session;
+
+  gnome_screen_cast_session =
+    gnome_screen_cast_create_session (gnome_screen_cast, NULL, error);
+  if (!gnome_screen_cast_session)
+    return FALSE;
+
+  screen_cast_session->gnome_screen_cast_session = gnome_screen_cast_session;
+
+  screen_cast_session->session_ready_handler_id =
+    g_signal_connect (gnome_screen_cast_session, "ready",
+                      G_CALLBACK (on_gnome_screen_cast_session_ready),
+                      screen_cast_session);
+  screen_cast_session->session_closed_handler_id =
+    g_signal_connect (gnome_screen_cast_session, "closed",
+                      G_CALLBACK (on_gnome_screen_cast_session_closed),
+                      screen_cast_session);
+
+  if (screen_cast_session->persist_mode != SCREEN_CAST_PERSIST_MODE_NONE)
+    screen_cast_session->streams_to_restore = g_ptr_array_ref (streams);
+
+  if (!gnome_screen_cast_session_record_selections (gnome_screen_cast_session,
+                                                    streams,
+                                                    &screen_cast_session->select,
+                                                    error))
+    return FALSE;
+
+  if (!gnome_screen_cast_session_start (gnome_screen_cast_session, error))
+    return FALSE;
+
+  return TRUE;
+}
+
+static void
+on_screen_cast_dialog_done_cb (GtkWidget              *widget,
+                               int                     dialog_response,
+                               ScreenCastPersistMode   persist_mode,
+                               GPtrArray              *streams,
+                               ScreenCastDialogHandle *dialog_handle)
 {
   int response;
 
@@ -135,7 +305,8 @@ screen_cast_dialog_done (GtkWidget *widget,
     {
     default:
       g_warning ("Unexpected response: %d", dialog_response);
-      /* Fall through */
+      G_GNUC_FALLTHROUGH;
+
     case GTK_RESPONSE_DELETE_EVENT:
       response = 2;
       break;
@@ -151,9 +322,13 @@ screen_cast_dialog_done (GtkWidget *widget,
 
   if (response == 0)
     {
+      ScreenCastSession *screen_cast_session = dialog_handle->session;
       g_autoptr(GError) error = NULL;
 
-      if (!start_session (dialog_handle->session, selections, &error))
+      screen_cast_session->persist_mode = MIN (screen_cast_session->persist_mode,
+                                               persist_mode);
+
+      if (!start_session (dialog_handle->session, streams, &error))
         {
           g_warning ("Failed to start session: %s", error->message);
           response = 2;
@@ -168,10 +343,10 @@ screen_cast_dialog_done (GtkWidget *widget,
 }
 
 static ScreenCastDialogHandle *
-create_screen_cast_dialog (ScreenCastSession *session,
+create_screen_cast_dialog (ScreenCastSession     *session,
                            GDBusMethodInvocation *invocation,
-                           Request *request,
-                           const char *parent_window)
+                           Request               *request,
+                           const char            *parent_window)
 {
   ScreenCastDialogHandle *dialog_handle;
   ExternalWindow *external_parent;
@@ -202,7 +377,8 @@ create_screen_cast_dialog (ScreenCastSession *session,
   g_object_ref_sink (fake_parent);
 
   dialog = GTK_WIDGET (screen_cast_dialog_new (request->app_id,
-                                               &session->select));
+                                               &session->select,
+                                               session->persist_mode));
   gtk_window_set_transient_for (GTK_WINDOW (dialog), GTK_WINDOW (fake_parent));
   gtk_window_set_modal (GTK_WINDOW (dialog), TRUE);
 
@@ -213,10 +389,9 @@ create_screen_cast_dialog (ScreenCastSession *session,
   dialog_handle->dialog = dialog;
 
   g_signal_connect (request, "handle-close",
-                    G_CALLBACK (handle_close), dialog_handle);
-
+                    G_CALLBACK (on_request_handle_close_cb), dialog_handle);
   g_signal_connect (dialog, "done",
-                    G_CALLBACK (screen_cast_dialog_done), dialog_handle);
+                    G_CALLBACK (on_screen_cast_dialog_done_cb), dialog_handle);
 
   gtk_widget_realize (dialog);
 
@@ -229,65 +404,250 @@ create_screen_cast_dialog (ScreenCastSession *session,
   return dialog_handle;
 }
 
-static ScreenCastSession *
-screen_cast_session_new (const char *app_id,
-                         const char *session_handle)
+static Monitor *
+find_monitor_by_string (const char *monitor_string)
 {
-  ScreenCastSession *screen_cast_session;
+  DisplayStateTracker *display_state_tracker = display_state_tracker_get ();
+  GList *l;
 
-  screen_cast_session = g_object_new (screen_cast_session_get_type (),
-                                      "id", session_handle,
-                                      NULL);
+  for (l = display_state_tracker_get_logical_monitors (display_state_tracker);
+       l;
+       l = l->next)
+    {
+      LogicalMonitor *logical_monitor = l->data;
+      GList *monitors;
 
-  return screen_cast_session;
+      for (monitors = logical_monitor_get_monitors (logical_monitor);
+           monitors;
+           monitors = monitors->next)
+        {
+          Monitor *monitor = monitors->data;
+
+          if (g_strcmp0 (monitor_get_match_string (monitor), monitor_string) == 0)
+            return monitor;
+        }
+    }
+
+  return NULL;
+}
+
+static Window *
+find_best_window_by_app_id_and_title (const char *app_id,
+                                      const char *title)
+{
+  ShellIntrospect *shell_introspect = shell_introspect_get ();
+  Window *best_match;
+  glong best_match_distance;
+  GList *l;
+
+  best_match = NULL;
+  best_match_distance = G_MAXLONG;
+
+  for (l = shell_introspect_get_windows (shell_introspect); l; l = l->next)
+    {
+      Window *window = l->data;
+      glong distance;
+
+      if (g_strcmp0 (window_get_app_id (window), app_id) != 0)
+        continue;
+
+      distance = str_distance (window_get_title (window), title);
+
+      if (distance == 0)
+        return window;
+
+      if (distance < best_match_distance)
+        {
+          best_match = window;
+          best_match_distance = distance;
+        }
+    }
+
+  /* If even the best match's window title is too different, don't
+   * restore it.
+   */
+  if (best_match_distance > strlen (title) / 2)
+    return NULL;
+
+  return best_match;
 }
 
 static gboolean
-handle_create_session (XdpImplScreenCast *object,
-                       GDBusMethodInvocation *invocation,
-                       const char *arg_handle,
-                       const char *arg_session_handle,
-                       const char *arg_app_id,
-                       GVariant *arg_options)
+restore_stream_from_data (ScreenCastSession *screen_cast_session)
+
 {
+  ScreenCastStreamInfo *info;
+  g_autoptr(GVariantIter) array_iter = NULL;
+  g_autoptr(GPtrArray) streams = NULL;
   g_autoptr(GError) error = NULL;
-  int response;
-  Session *session;
-  GVariantBuilder results_builder;
+  ScreenCastSourceType source_type;
+  GVariant *data;
+  uint32_t id;
+  int64_t creation_time;
+  int64_t last_used_time;
 
-  session = (Session *)screen_cast_session_new (arg_app_id,
-                                                arg_session_handle);
+  if (!screen_cast_session->restored.data)
+    return FALSE;
 
-  if (!session_export (session,
-                       g_dbus_method_invocation_get_connection (invocation),
-                       &error))
+  streams = g_ptr_array_new_with_free_func (g_free);
+
+  g_variant_get (screen_cast_session->restored.data,
+                 RESTORE_VARIANT_TYPE,
+                 &creation_time,
+                 &last_used_time,
+                 &array_iter);
+
+  while (g_variant_iter_next (array_iter, "(uuv)", &id, &source_type, &data))
     {
-      g_clear_object (&session);
-      g_warning ("Failed to create screen cast session: %s", error->message);
-      response = 2;
-      goto out;
+      switch (source_type)
+        {
+        case SCREEN_CAST_SOURCE_TYPE_MONITOR:
+          {
+            if (!(screen_cast_session->select.source_types & SCREEN_CAST_SOURCE_TYPE_MONITOR) ||
+                !g_variant_check_format_string (data, MONITOR_TYPE, FALSE))
+              goto fail;
+
+            const char *match_string = g_variant_get_string (data, NULL);
+            Monitor *monitor = find_monitor_by_string (match_string);
+
+            if (!monitor)
+              goto fail;
+
+            info = g_new0 (ScreenCastStreamInfo, 1);
+            info->type = SCREEN_CAST_SOURCE_TYPE_MONITOR;
+            info->data.monitor = monitor;
+            info->id = id;
+            g_ptr_array_add (streams, info);
+          }
+          break;
+
+        case SCREEN_CAST_SOURCE_TYPE_WINDOW:
+          {
+            if (!(screen_cast_session->select.source_types & SCREEN_CAST_SOURCE_TYPE_WINDOW) ||
+                !g_variant_check_format_string (data, WINDOW_TYPE, FALSE))
+              goto fail;
+
+            const char *app_id = NULL;
+            const char *title = NULL;
+            Window *window;
+
+            g_variant_get (data, "(&s&s)", &app_id, &title);
+
+            window = find_best_window_by_app_id_and_title (app_id, title);
+
+            if (!window)
+              goto fail;
+
+            info = g_new0 (ScreenCastStreamInfo, 1);
+            info->type = SCREEN_CAST_SOURCE_TYPE_WINDOW;
+            info->data.window = window;
+            info->id = id;
+            g_ptr_array_add (streams, info);
+          }
+          break;
+
+        case SCREEN_CAST_SOURCE_TYPE_VIRTUAL:
+          {
+            if (!(screen_cast_session->select.source_types & SCREEN_CAST_SOURCE_TYPE_VIRTUAL) ||
+                !g_variant_check_format_string (data, VIRTUAL_TYPE, FALSE))
+              goto fail;
+
+            info = g_new0 (ScreenCastStreamInfo, 1);
+            info->type = SCREEN_CAST_SOURCE_TYPE_VIRTUAL;
+            info->id = id;
+            g_ptr_array_add (streams, info);
+          }
+          break;
+
+        default:
+          goto fail;
+        }
     }
 
-  response = 0;
+  screen_cast_session->restored.creation_time = creation_time;
 
-out:
-  g_variant_builder_init (&results_builder, G_VARIANT_TYPE_VARDICT);
-  xdp_impl_screen_cast_complete_create_session (object,
-                                                invocation,
-                                                response,
-                                                g_variant_builder_end (&results_builder));
+  start_session (screen_cast_session, streams, &error);
+
+  if (error)
+    {
+      g_warning ("Error restoring stream from session: %s", error->message);
+      return FALSE;
+    }
+
+  return TRUE;
+
+fail:
+  return FALSE;
+}
+
+static gboolean
+handle_start (XdpImplScreenCast     *object,
+              GDBusMethodInvocation *invocation,
+              const char            *arg_handle,
+              const char            *arg_session_handle,
+              const char            *arg_app_id,
+              const char            *arg_parent_window,
+              GVariant              *arg_options)
+{
+  const char *sender;
+  g_autoptr(Request) request = NULL;
+  ScreenCastSession *screen_cast_session;
+  GVariantBuilder results_builder;
+
+  sender = g_dbus_method_invocation_get_sender (invocation);
+  request = request_new (sender, arg_app_id, arg_handle);
+  request_export (request,
+                  g_dbus_method_invocation_get_connection (invocation));
+
+  screen_cast_session =
+    (ScreenCastSession *)lookup_session (arg_session_handle);
+  if (!screen_cast_session)
+    {
+      g_warning ("Attempted to start non existing screen cast session");
+      goto err;
+    }
+
+  if (screen_cast_session->dialog_handle)
+    {
+      g_warning ("Screen cast dialog already open");
+      goto err;
+    }
+
+  screen_cast_session->start_invocation = invocation;
+
+  if (!restore_stream_from_data (screen_cast_session))
+    {
+      ScreenCastDialogHandle *dialog_handle;
+
+      dialog_handle = create_screen_cast_dialog (screen_cast_session,
+                                                 invocation,
+                                                 request,
+                                                 arg_parent_window);
+
+      screen_cast_session->dialog_handle = dialog_handle;
+    }
+
+  return TRUE;
+
+err:
+  g_variant_builder_init (&results_builder, G_VARIANT_TYPE ("a(ua{sv}"));
+  xdp_impl_screen_cast_complete_start (object, invocation, 2,
+                                       g_variant_builder_end (&results_builder));
 
   return TRUE;
 }
 
 static gboolean
-handle_select_sources (XdpImplScreenCast *object,
+handle_select_sources (XdpImplScreenCast     *object,
                        GDBusMethodInvocation *invocation,
-                       const char *arg_handle,
-                       const char *arg_session_handle,
-                       const char *arg_app_id,
-                       GVariant *arg_options)
+                       const char            *arg_handle,
+                       const char            *arg_session_handle,
+                       const char            *arg_app_id,
+                       GVariant              *arg_options)
 {
+  g_autofree gchar *provider = NULL;
+  g_autoptr(GVariant) restore_data = NULL;
+  ScreenCastSession *screen_cast_session;
   Session *session;
   int response;
   uint32_t types;
@@ -296,6 +656,7 @@ handle_select_sources (XdpImplScreenCast *object,
   ScreenCastSelection select;
   GVariantBuilder results_builder;
   GVariant *results;
+  uint32_t version;
 
   session = lookup_session (arg_session_handle);
   if (!session)
@@ -359,6 +720,21 @@ handle_select_sources (XdpImplScreenCast *object,
       response = 2;
     }
 
+  screen_cast_session = (ScreenCastSession *)session;
+  g_variant_lookup (arg_options, "persist_mode", "u", &screen_cast_session->persist_mode);
+
+  if (g_variant_lookup (arg_options, "restore_data", "(suv)", &provider, &version, &restore_data))
+    {
+      if (!g_variant_check_format_string (restore_data, RESTORE_VARIANT_TYPE, FALSE))
+        {
+          g_warning ("Cannot parse restore data, ignoring");
+          goto out;
+        }
+
+      if (g_strcmp0 (provider, "GNOME") == 0 && version == RESTORE_FORMAT_VERSION)
+        screen_cast_session->restored.data = g_variant_ref (restore_data);
+    }
+
 out:
   g_variant_builder_init (&results_builder, G_VARIANT_TYPE_VARDICT);
   results = g_variant_builder_end (&results_builder);
@@ -368,144 +744,41 @@ out:
   return TRUE;
 }
 
-static void
-start_done (ScreenCastSession *screen_cast_session)
-{
-  GnomeScreenCastSession *gnome_screen_cast_session;
-  GVariantBuilder streams_builder;
-  GVariantBuilder results_builder;
-
-  g_variant_builder_init (&results_builder, G_VARIANT_TYPE_VARDICT);
-  g_variant_builder_init (&streams_builder, G_VARIANT_TYPE ("a(ua{sv})"));
-
-  gnome_screen_cast_session = screen_cast_session->gnome_screen_cast_session;
-  gnome_screen_cast_session_add_stream_properties (gnome_screen_cast_session,
-                                                   &streams_builder);
-
-  g_variant_builder_add (&results_builder, "{sv}",
-                         "streams",
-                         g_variant_builder_end (&streams_builder));
-
-  xdp_impl_screen_cast_complete_start (XDP_IMPL_SCREEN_CAST (impl),
-                                       screen_cast_session->start_invocation, 0,
-                                       g_variant_builder_end (&results_builder));
-  screen_cast_session->start_invocation = NULL;
-}
-
-static void
-on_gnome_screen_cast_session_ready (GnomeScreenCastSession *gnome_screen_cast_session,
-                                    ScreenCastSession *screen_cast_session)
-{
-  start_done (screen_cast_session);
-}
-
-static void
-on_gnome_screen_cast_session_closed (GnomeScreenCastSession *gnome_screen_cast_session,
-                                     ScreenCastSession *screen_cast_session)
-{
-  session_close ((Session *)screen_cast_session);
-}
-
 static gboolean
-start_session (ScreenCastSession *screen_cast_session,
-               GVariant *selections,
-               GError **error)
-{
-  GnomeScreenCastSession *gnome_screen_cast_session;
-  g_autoptr(GVariant) source_selections = NULL;
-
-  gnome_screen_cast_session =
-    gnome_screen_cast_create_session (gnome_screen_cast, NULL, error);
-  if (!gnome_screen_cast_session)
-    return FALSE;
-
-  screen_cast_session->gnome_screen_cast_session = gnome_screen_cast_session;
-
-  screen_cast_session->session_ready_handler_id =
-    g_signal_connect (gnome_screen_cast_session, "ready",
-                      G_CALLBACK (on_gnome_screen_cast_session_ready),
-                      screen_cast_session);
-  screen_cast_session->session_closed_handler_id =
-    g_signal_connect (gnome_screen_cast_session, "closed",
-                      G_CALLBACK (on_gnome_screen_cast_session_closed),
-                      screen_cast_session);
-
-  g_variant_lookup (selections, "selected_screen_cast_sources", "@a(u?)",
-                    &source_selections);
-  if (!gnome_screen_cast_session_record_selections (gnome_screen_cast_session,
-                                                    source_selections,
-                                                    &screen_cast_session->select,
-                                                    error))
-    return FALSE;
-
-  if (!gnome_screen_cast_session_start (gnome_screen_cast_session, error))
-    return FALSE;
-
-  return TRUE;
-}
-
-static void
-cancel_start_session (ScreenCastSession *screen_cast_session,
-                      int response)
+handle_create_session (XdpImplScreenCast     *object,
+                       GDBusMethodInvocation *invocation,
+                       const char            *arg_handle,
+                       const char            *arg_session_handle,
+                       const char            *arg_app_id,
+                       GVariant              *arg_options)
 {
+  g_autoptr(GError) error = NULL;
+  int response;
+  Session *session;
   GVariantBuilder results_builder;
 
-  g_variant_builder_init (&results_builder, G_VARIANT_TYPE_VARDICT);
-  xdp_impl_screen_cast_complete_start (XDP_IMPL_SCREEN_CAST (impl),
-                                       screen_cast_session->start_invocation,
-                                       response,
-                                       g_variant_builder_end (&results_builder));
-}
+  session = g_object_new (screen_cast_session_get_type (),
+                          "id", arg_session_handle,
+                          NULL);
 
-static gboolean
-handle_start (XdpImplScreenCast *object,
-              GDBusMethodInvocation *invocation,
-              const char *arg_handle,
-              const char *arg_session_handle,
-              const char *arg_app_id,
-              const char *arg_parent_window,
-              GVariant *arg_options)
-{
-  const char *sender;
-  g_autoptr(Request) request = NULL;
-  ScreenCastSession *screen_cast_session;
-  ScreenCastDialogHandle *dialog_handle;
-  GVariantBuilder results_builder;
-
-  sender = g_dbus_method_invocation_get_sender (invocation);
-  request = request_new (sender, arg_app_id, arg_handle);
-  request_export (request,
-                  g_dbus_method_invocation_get_connection (invocation));
-
-  screen_cast_session =
-    (ScreenCastSession *)lookup_session (arg_session_handle);
-  if (!screen_cast_session)
+  if (!session_export (session,
+                       g_dbus_method_invocation_get_connection (invocation),
+                       &error))
     {
-      g_warning ("Attempted to start non existing screen cast session");
-      goto err;
+      g_clear_object (&session);
+      g_warning ("Failed to create screen cast session: %s", error->message);
+      response = 2;
+      goto out;
     }
 
-  if (screen_cast_session->dialog_handle)
-    {
-      g_warning ("Screen cast dialog already open");
-      goto err;
-    }
+  response = 0;
 
-  dialog_handle = create_screen_cast_dialog (screen_cast_session,
-                                             invocation,
-                                             request,
-                                             arg_parent_window);
-
-
-  screen_cast_session->start_invocation = invocation;
-  screen_cast_session->dialog_handle = dialog_handle;
-
-  return TRUE;
-
-err:
-  g_variant_builder_init (&results_builder, G_VARIANT_TYPE ("a(ua{sv}"));
-  xdp_impl_screen_cast_complete_start (object, invocation, 2,
-                                       g_variant_builder_end (&results_builder));
+out:
+  g_variant_builder_init (&results_builder, G_VARIANT_TYPE_VARDICT);
+  xdp_impl_screen_cast_complete_create_session (object,
+                                                invocation,
+                                                response,
+                                                g_variant_builder_end (&results_builder));
 
   return TRUE;
 }
@@ -538,9 +811,12 @@ on_gnome_screen_cast_enabled (GnomeScreenCast *gnome_screen_cast)
 
   available_cursor_modes = SCREEN_CAST_CURSOR_MODE_NONE;
   if (gnome_api_version >= 2)
-    available_cursor_modes |= (SCREEN_CAST_CURSOR_MODE_HIDDEN |
-                               SCREEN_CAST_CURSOR_MODE_EMBEDDED |
-                               SCREEN_CAST_CURSOR_MODE_METADATA);
+    {
+      available_cursor_modes |= SCREEN_CAST_CURSOR_MODE_HIDDEN |
+                                SCREEN_CAST_CURSOR_MODE_EMBEDDED |
+                                SCREEN_CAST_CURSOR_MODE_METADATA;
+    }
+
   g_object_set (G_OBJECT (impl),
                 "available-cursor-modes", available_cursor_modes,
                 NULL);
@@ -572,21 +848,6 @@ on_gnome_screen_cast_disabled (GDBusConnection *connection,
     }
 }
 
-gboolean
-screen_cast_init (GDBusConnection *connection,
-                  GError **error)
-{
-  impl_connection = connection;
-  gnome_screen_cast = gnome_screen_cast_new (connection);
-
-  g_signal_connect (gnome_screen_cast, "enabled",
-                    G_CALLBACK (on_gnome_screen_cast_enabled), NULL);
-  g_signal_connect (gnome_screen_cast, "disabled",
-                    G_CALLBACK (on_gnome_screen_cast_disabled), NULL);
-
-  return TRUE;
-}
-
 static void
 screen_cast_session_close (Session *session)
 {
@@ -614,6 +875,8 @@ screen_cast_session_finalize (GObject *object)
 {
   ScreenCastSession *screen_cast_session = (ScreenCastSession *)object;
 
+  g_clear_pointer (&screen_cast_session->streams_to_restore, g_ptr_array_unref);
+  g_clear_pointer (&screen_cast_session->restored.data, g_variant_unref);
   g_clear_object (&screen_cast_session->gnome_screen_cast_session);
 
   G_OBJECT_CLASS (screen_cast_session_parent_class)->finalize (object);
@@ -622,6 +885,8 @@ screen_cast_session_finalize (GObject *object)
 static void
 screen_cast_session_init (ScreenCastSession *screen_cast_session)
 {
+  screen_cast_session->persist_mode = SCREEN_CAST_PERSIST_MODE_NONE;
+  screen_cast_session->restored.creation_time = -1;
 }
 
 static void
@@ -636,3 +901,26 @@ screen_cast_session_class_init (ScreenCastSessionClass *klass)
   session_class = (SessionClass *)klass;
   session_class->close = screen_cast_session_close;
 }
+
+gboolean
+screen_cast_init (GDBusConnection  *connection,
+                  GError          **error)
+{
+  /*
+   * Ensure ShellIntrospect and DisplayStateTracker are initialized before
+   * any screencast session is created to avoid race conditions when restoring
+   * previous streams.
+   */
+  display_state_tracker_get ();
+  shell_introspect_get ();
+
+  impl_connection = connection;
+  gnome_screen_cast = gnome_screen_cast_new (connection);
+
+  g_signal_connect (gnome_screen_cast, "enabled",
+                    G_CALLBACK (on_gnome_screen_cast_enabled), NULL);
+  g_signal_connect (gnome_screen_cast, "disabled",
+                    G_CALLBACK (on_gnome_screen_cast_disabled), NULL);
+
+  return TRUE;
+}
