@@ -31,7 +31,7 @@
 #include <sys/stat.h>
 #include <fcntl.h>
 
-#include <gtk/gtk.h>
+#include <adwaita.h>
 
 #include <gio/gio.h>
 #include <gio/gdesktopappinfo.h>
@@ -42,9 +42,12 @@
 
 #include "xdg-desktop-portal-dbus.h"
 
+#include "account.h"
 #include "appchooser.h"
 #include "background.h"
+#include "filechooser.h"
 #include "lockdown.h"
+#include "print.h"
 #include "screenshot.h"
 #include "screencast.h"
 #include "remotedesktop.h"
@@ -102,6 +105,18 @@ on_bus_acquired (GDBusConnection *connection,
 {
   GError *error = NULL;
 
+  if (!file_chooser_init (connection, &error))
+    {
+      g_warning ("error: %s\n", error->message);
+      g_clear_error (&error);
+    }
+
+  if (!account_init (connection, &error))
+    {
+      g_warning ("error: %s\n", error->message);
+      g_clear_error (&error);
+    }
+
   if (!app_chooser_init (connection, &error))
     {
       g_warning ("error: %s\n", error->message);
@@ -120,6 +135,12 @@ on_bus_acquired (GDBusConnection *connection,
       g_clear_error (&error);
     }
 
+  if (!print_init (connection, &error))
+    {
+      g_warning ("error: %s\n", error->message);
+      g_clear_error (&error);
+    }
+
   if (!screenshot_init (connection, &error))
     {
       g_warning ("error: %s\n", error->message);
@@ -183,6 +204,8 @@ main (int argc, char *argv[])
 
   /* Avoid pointless and confusing recursion */
   g_unsetenv ("GTK_USE_PORTAL");
+  g_setenv ("ADW_DISABLE_PORTAL", "1", TRUE);
+  g_setenv ("GSK_RENDERER", "cairo", TRUE);
 
   gtk_init ();
 
@@ -196,7 +219,7 @@ main (int argc, char *argv[])
       "Documentation for the available D-Bus interfaces can be found at\n"
       "https://flatpak.github.io/xdg-desktop-portal/portal-docs.html\n"
       "\n"
-      "Please report issues at https://gitlab.gnome.org/feaneron/xdg-desktop-portal-gnome/issues");
+      "Please report issues at https://gitlab.gnome.org/GNOME/xdg-desktop-portal-gnome/issues");
   g_option_context_add_main_entries (context, entries, NULL);
   if (!g_option_context_parse (context, &argc, &argv, &error))
     {
@@ -241,6 +264,8 @@ main (int argc, char *argv[])
                              NULL,
                              NULL);
 
+  adw_init ();
+
   g_main_loop_run (loop);
 
   g_bus_unown_name (owner_id);
